/*

             **************************************************
             ****               VEGA - Solvate             ****
             ****             Solvate a molecule           ****
             **** Copyright 1996-2000, Alessandro Pedretti ****
             **************************************************


*/


#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "globdef.h"
#include "globvar.h"
#include "formats.h"

#define  COVTOL			2.5     /* Overlapping of covalent radius */


/**** Structures ****/

typedef struct {
  float         Gx, Gy, Gz;             /* Coordinates of geometry center */
  float         MaxX, MaxY, MaxZ;       /* Molecule extension             */
} MOLINFO;


/**** Prototypes ****/

void CalcMolInfo(ATOMO *, ULONG, MOLINFO *);


/**** Solvate a molecule ****/

BOOL Solvate(char *ShellFile, ATOMO *InizAtm, ULONG *TotAtm)
{
  ATOMO                 *BegRes, *InizSol, *LastAtm, *PrecAtm, *PrecRes;
  FILE                  *FH;
  RECORD                Buf;
  float                 *InizCovRad, *InizCovRadSol;
  float                 *CovRad;
  float                 Mx, My, Mz;
  float                 Nx, Ny, Nz;
  MOLINFO               MolInfo, SolInfo;
  register ATOMO         *Atm, *Sol;
  ULONG                 j, k;

  BOOL			Ret    = TRUE;
  float			Rad    = Prefs.SOL_RAD;
  ULONG			TotSol = 0;


  if (FH = fopen(ShellFile, "r")) {
    if (InizCovRad = (float *)AssignCovRad(InizAtm, *TotAtm)) {
      *Buf.Line = 0;
      PrintProg("* Solvating the molecule\n" \
                "  Reading shell file");
      if (InizSol = (ATOMO *)PDBLoad(FH, &Buf, &TotSol)) {

        /**** Calculate dimensions and geometry center ****/

        CalcMolInfo(InizAtm, *TotAtm, &MolInfo);
        CalcMolInfo(InizSol, TotSol, &SolInfo);

        if ((MolInfo.MaxX <= SolInfo.MaxX) &&
            (MolInfo.MaxY <= SolInfo.MaxY) &&
            (MolInfo.MaxZ <= SolInfo.MaxZ)) {

          /**** Translate the cluster ****/

          Mx = MolInfo.Gx - SolInfo.Gx;
          My = MolInfo.Gy - SolInfo.Gy;
          Mz = MolInfo.Gz - SolInfo.Gz;
          for(Atm = InizSol; Atm; Atm = Atm -> Ptr) {
            Atm -> x += Mx;
            Atm -> y += My;
            Atm -> z += Mz;
          }

          /**** Cut the shell ****/

          switch(Prefs.SOL_SHAPE) {

          case SOL_SHAPE_BOX:
            Rad /= 2.0;
            Mx = MolInfo.Gx + Rad;
            My = MolInfo.Gy + Rad;
            Mz = MolInfo.Gz + Rad;
            Nx = MolInfo.Gx - Rad;
            Ny = MolInfo.Gy - Rad;
            Nz = MolInfo.Gz - Rad;
            break;

          case SOL_SHAPE_SPHERE:
            Rad *= Rad;
            Mx   = MolInfo.Gx;
            My   = MolInfo.Gy;
            Mz   = MolInfo.Gz;
            break;

          case SOL_SHAPE_LAYER:
            Rad *= Rad;
            break;
          }

          BegRes  = InizSol;
          PrecRes = InizSol;
          Sol     = InizSol;
          do {
            if (BegRes -> ResSeq.L != Sol -> ResSeq.L) {
              PrecRes = PrecAtm;
              BegRes  = Sol;
            }
            switch(Prefs.SOL_SHAPE) {
            case SOL_SHAPE_BOX:
              if ((Sol -> x > Mx) || (Sol -> x < Nx) ||
                  (Sol -> y > My) || (Sol -> y < Ny) ||
                  (Sol -> z > Mz) || (Sol -> z < Nz)) {
                RemoveRes(&BegRes, &PrecRes, &InizSol, &TotSol);
                Sol = BegRes;
              } else {
                PrecAtm = Sol;
                Sol = Sol -> Ptr;
              }
              break;

            case SOL_SHAPE_SPHERE:
              if ((Quad(Sol -> x - Mx) + Quad(Sol -> y - My) + Quad(Sol -> z - Mz)) > Rad) {
                RemoveRes(&BegRes, &PrecRes, &InizSol, &TotSol);
                Sol = BegRes;
              } else {
                PrecAtm = Sol;
                Sol = Sol -> Ptr;
              }
              break;

            case SOL_SHAPE_LAYER:
              k = TRUE;
              for(Atm = InizAtm; (Atm) && (k); Atm = Atm -> Ptr) {
                if (QuadDist(Atm, Sol) <= Rad) {
                  do {
                    BegRes -> Flags |= ATM_FLG_GENERIC;
                    BegRes = BegRes -> Ptr;
                  } while((BegRes) && (BegRes -> ResSeq.L == Atm -> ResSeq.L));
                  Sol = BegRes;
                  k = FALSE;
                }
              } /* End of atom loop */
              if (k) {
                PrecAtm = Sol;
                Sol = Sol -> Ptr;
              }
              break;
            }
          } while(Sol); /* End of solvent loop */

          /**** Remove layer atoms ****/

          if (Prefs.SOL_SHAPE == SOL_SHAPE_LAYER) {
            PrecAtm = InizSol;
            Sol     = InizSol;
            do {
              if (Sol -> Flags & ATM_FLG_GENERIC) {
                Sol -> Flags &= ATM_FLG_GENERIC - 1;
                PrecAtm = Sol;
                Sol     = Sol -> Ptr;
              } else {
                LastAtm = Sol; /* Sol is a register pointer ! */
                RemoveAtm(&LastAtm, &PrecAtm, &InizSol);
                Sol = LastAtm;
                --TotSol;
              }
            } while(Sol); /* End of solvent loop */
          }

          /**** Removing bumping solvent molecules ****/

          if ((TotSol) && (InizCovRadSol = (float *)AssignCovRad(InizSol, TotSol))) {
            k = 1;
            for(Atm = InizSol; Atm; Atm = Atm -> Ptr) Atm -> Num = ++k;
            for(Atm = InizAtm; Atm; Atm = Atm -> Ptr) {
              CovRad  = InizCovRad;
              BegRes  = InizSol;
              PrecRes = InizSol;
              Sol     = InizSol;
              do {
                if (BegRes -> ResSeq.L != Sol -> ResSeq.L) {
                  PrecRes = PrecAtm;
                  BegRes  = Sol;
                }
                if (QuadDist(Atm, Sol) < Quad(*CovRad + InizCovRadSol[Sol -> Num - 1])) {
                  RemoveRes(&BegRes, &PrecRes, &InizSol, &TotSol);
                  Sol = BegRes;
                } else {
                  PrecAtm = Sol;
                  Sol = Sol -> Ptr;
                }
              } while(Sol); /* End of solvent loop */
              ++CovRad;
            } /* End of molecule loop */
            FREE(InizCovRadSol);
          } else Ret = FALSE;

          /**** Make the binding between molecule and solvent ****/

          if ((Ret) && (InizSol)) {

            /**** If needed, build the solvent connectivity ****/

            if (GlobSwitch[3]) Ret = Connect(InizSol, TotSol);
            for(Atm = InizAtm; Atm; Atm = Atm -> Ptr)
              LastAtm = Atm;
            *TotAtm += TotSol;
            LastAtm -> Ptr = InizSol;

            /**** Renumber the atoms ****/

            j      = 1;
            k      = LastAtm -> Num;
            BegRes = InizSol;
            for(Atm = InizSol; Atm; Atm = Atm -> Ptr) {
              if (BegRes -> ResSeq.L != Atm -> ResSeq.L) {
                BegRes = Atm;
                ++j;
              }
              Atm -> Num = ++k;
            } /* End of solvent loop */
            if (!GlobSwitch[4]) printf("  Added %d solvent molecules\n", j);
            Prefs.RENSTART = 1;
            RenRes(InizAtm);
          }
        } else Ret = Errore("The molecule exceeds the shell dimensions");
        if (!Ret) PDBFree(InizSol);
      } else Ret = Errore("while reading shell file");
      FREE(InizCovRad);
    } else Ret = FALSE;
    fclose(FH);
  } else Ret = Errore("Can't open shell file");

  return Ret;
}


/**** Calculates geometry center and max dimensions ****/

void CalcMolInfo(register ATOMO *Atm, ULONG TotAtm, register MOLINFO *Info)
{
  float         MinX, MinY, MinZ;

  Info -> Gx   = 0.0;
  Info -> Gy   = 0.0;
  Info -> Gz   = 0.0;
  Info -> MaxX = Atm -> x;
  Info -> MaxY = Atm -> y;
  Info -> MaxZ = Atm -> z;
  MinX         = Atm -> x;
  MinY         = Atm -> y;
  MinZ         = Atm -> z;
  while(Atm) {
    if (Atm -> x > Info -> MaxX) Info -> MaxX = Atm -> x;
    if (Atm -> x < MinX) MinX = Atm -> x;
    if (Atm -> y > Info -> MaxY) Info -> MaxY = Atm -> y;
    if (Atm -> y < MinY) MinY = Atm -> y;
    if (Atm -> z > Info -> MaxZ) Info -> MaxZ = Atm -> z;
    if (Atm -> z < MinZ) MinZ = Atm -> z;
    Info -> Gx += Atm -> x;
    Info -> Gy += Atm -> y;
    Info -> Gz += Atm -> z;
    Atm         = Atm -> Ptr;
  } /* End of atom loop */
  Info -> Gx   /= TotAtm;
  Info -> Gy   /= TotAtm;
  Info -> Gz   /= TotAtm;
  Info -> MaxX  = fabs(Info -> MaxX - MinX) + COVTOL;
  Info -> MaxY  = fabs(Info -> MaxY - MinY) + COVTOL;
  Info -> MaxZ  = fabs(Info -> MaxZ - MinZ) + COVTOL;
}

